from pathlib import Path
from typing import List, Optional

from motllo.ops import Folder, tree, tree_links
from motllo.traverser import Traverser


def gitignore(path: Path, for_path=True):
    """Parse a gitignore to keep pathlib happy"""
    if not path.exists():
        return []
    globs = []
    with path.open() as gitignore_file:
        lines = gitignore_file.read().split("\n")
    clean_globs = [l for l in lines if not l.startswith("#") and l != ""]
    globs = [f"{g}/*" if g.endswith("**") else g for g in clean_globs]
    if for_path:
        globs = [f"{path.parent}/{g}" for g in globs]
    return globs


def global_gitignore():
    """Parse global gitignore"""
    return gitignore(Path.home() / ".gitignore_global", for_path=False)


def project_gitignore(path: Path):
    """Parse all project gitignores (including in folders)"""
    ignores = []
    for item in path.iterdir():
        if item.name == ".gitignore":
            ignores += gitignore(item, for_path=True)
        if item.is_dir():
            ignores += project_gitignore(item)
    return ignores


def full_gitignore(path: Path):
    """Combined gitignores"""
    return project_gitignore(path) + global_gitignore()


SUFFIX_TO_LANG = {
    "py": "python",
    "sbt": "scala",
    "scala": "scala",
    "toml": "toml",
    "json": "json",
    "lock": "toml",
    "yaml": "yaml",
}

LANG_TO_ELLIPSIS = {
    "python": "# ...",
    "scala": "// ...",
    "toml": "[...]",
    "json": "...",
}


def language(suffix):
    """Get language for Markdown code block"""
    return SUFFIX_TO_LANG.get(suffix, "")


def ellipsis(suffix):
    """Comment-level ellipsis for the files"""
    return LANG_TO_ELLIPSIS.get(language(suffix), "...")


def build_file_markdown(current: Folder, base="", max_length=15):
    """Generate markdown out of the folder structure"""
    markdown: List[str] = []
    for item in current.iterdir():
        if item.is_dir():
            markdown += build_file_markdown(
                item, base=base + "/" + item.basename, max_length=max_length
            )
        else:
            if base.startswith("/"):
                base = base[1:]
            if base == "":
                basename = f"{item.name}"
            else:
                basename = f"{base}/{item.name}"
            markdown += [""] + [f"# `{basename}`"]
            if item.contents is not None:
                markdown += [""] + [f"```{language(item.suffix)}"]
                if item.suffix != "md":
                    if len(item.contents) > max_length >= 0:
                        markdown += (
                            item.contents[0 : max_length - 1]
                            + [""]
                            + [ellipsis(item.suffix)]
                        )
                    else:
                        markdown += item.contents
                else:
                    markdown += [
                        "Content from Markdown files is ignored, since the output would break parsing"
                    ]
                markdown += ["```"]
    return markdown


def build_tree(
    path: Path, ignore_globs: Optional[List[str]], include_globs: Optional[List[str]]
):
    """Build the tree from a path, given a glob"""
    structure = Traverser(ignore_globs=ignore_globs, include_globs=include_globs)(path)
    structure = structure.prune()
    return structure


def text_tree(
    path: Path,
    ignore_globs: Optional[List[str]] = None,
    include_globs: Optional[List[str]] = None,
):
    """Generate the textual tree representation only"""
    structure = build_tree(path, ignore_globs, include_globs)
    return "\n".join(list(tree(structure)))


def build_markdown(structure: Folder, max_length):
    """Generate markdown from a path, given ignore files"""
    built_tree = tree(structure)
    markdown = [""] + ["# Tree structure"]
    markdown += [""] + ["```"] + list(built_tree) + ["```"] + [""]
    markdown += list(tree_links(structure))
    markdown += [""] + build_file_markdown(structure, base="", max_length=max_length)
    return markdown


def write_markdown(markdown: List[str], path: Path):
    """Write final markdown to a path"""
    with path.open("w") as destination:
        for line in markdown:
            destination.write(line)
            destination.write("\n")



import logging
import sys
from pathlib import Path

from motllo.markdown_parser import (MARKDOWN_PARSER, BareCodeBlock, CodeBlock,
                                    HeadingBlock, ListBlock)
from motllo.ops import Folder
from motllo.tree_parser import TreeParser

logger = logging.getLogger("motllo.read_markdown")

TREE_KEY = "Markdown Tree Structure"


def structure_filler(folder, file_contents, file_replacements):
    """Populate a folder tree structure with file contents"""
    for item in folder.iterdir():
        if item.is_dir():
            structure_filler(item, file_contents, file_replacements)
        else:
            if item.as_posix().startswith("/."):
                cleaned = item.as_posix().replace("/./", "/")
            else:
                cleaned = item.as_posix()
            if cleaned not in file_contents.keys():
                raise Exception(f"Damn file not there {cleaned}")
            item.set_contents(file_contents[cleaned])
            if cleaned not in file_replacements:
                logger.debug("No replacements found for marker %s", cleaned)
            else:
                item.set_replacements(file_replacements[cleaned])


def replace_replacements(file_replacement, replacements, to_be_replaced):
    """Apply replacements to file identifier"""
    replaced = to_be_replaced
    common_keys = [key for key in file_replacement if key in replacements]
    for key in common_keys:
        old_value = file_replacement[key]
        new_value = replacements[key]
        replaced = replaced.replace(old_value, new_value)
    return replaced


def replace_file_contents(replacements, file_contents, file_replacements):
    """Modify tree structure according to the replacements provided"""
    replaced_file_contents = {}
    for marker in file_contents:
        logger.debug("File marker %s", marker)
        contents = file_contents[marker]
        if (
            replacements is not None
            and file_replacements.get(TREE_KEY, None) is not None
        ):
            new_file_marker = replace_replacements(
                file_replacements[TREE_KEY], replacements, marker
            )
            if new_file_marker != marker:
                logger.info(
                    "Replaced file reference `%s` with `%s`", marker, new_file_marker
                )
            replaced_file_contents[new_file_marker] = contents
        else:
            replaced_file_contents[marker] = contents
    return replaced_file_contents


def _is_heading_block(item, with_title=None):
    if with_title is not None:
        return (
            isinstance(item, HeadingBlock) and item.text.strip().lower() == with_title
        )
    return isinstance(item, HeadingBlock)


def process_markdown_definitions(markdown, replacements):
    """Process markdown, finding replacements needed"""
    file_marker = None
    replacement_marker = None
    file_contents = {}
    file_replacements = {}
    required_keys = {}
    for item in markdown:
        if _is_heading_block(item):
            if _is_heading_block(item, with_title="tree structure"):
                file_marker = TREE_KEY
            elif _is_heading_block(item, with_title="replacements"):
                replacement_marker = file_marker
            else:
                filename = item.text.strip()
                file_marker = filename.replace("`", "")
        elif isinstance(item, ListBlock) and replacement_marker is not None:
            if ":" not in item.text:
                continue
            logger.info(
                "Found a replacement block %s for section `%s`",
                item.text,
                replacement_marker,
            )
            key, in_code = item.text.split(":")
            in_code = in_code.replace("`", "")
            key = key.replace("`", "")
            required_keys[key] = True
            replacer = {key: in_code.strip()}
            logger.debug("Replacer: %s", replacer)
            if replacement_marker not in file_replacements:
                file_replacements[replacement_marker] = replacer
            else:
                file_replacements[replacement_marker].update(replacer)
        elif isinstance(item, (CodeBlock, BareCodeBlock)):
            contents = item.text.strip()
            if file_marker not in file_contents:
                file_contents[file_marker] = contents
            else:
                file_contents[file_marker] += "\n\n" + contents
    for key in required_keys:
        if key not in replacements:
            logger.warning(
                "You have provided no replacement for `%s` in the CLI (use the -r flag)",
                key,
            )
    return file_contents, file_replacements


def parse_markdown_to_structure(markdown, replacements):
    """Convert a parsed Markdown document into a folder structure"""
    file_contents, file_replacements = process_markdown_definitions(
        markdown, replacements
    )

    replaced_file_contents = replace_file_contents(
        replacements, file_contents, file_replacements
    )

    if TREE_KEY not in replaced_file_contents:
        logger.error(
            "Tree structure section not found in the Markdown document. This is required"
        )
        sys.exit(-1)
    tree_lines = [
        l for l in replaced_file_contents[TREE_KEY].split("\n") if l != ""
    ]  # This is pretty crappy
    replaced_tree = []
    replaced_file_replacements = {}
    if replacements is not None and file_replacements.get(TREE_KEY, None) is not None:
        tree_replacements = file_replacements[TREE_KEY]
        for line in tree_lines:
            replaced_tree += [
                replace_replacements(tree_replacements, replacements, line)
            ]
        for key in file_replacements:
            replaced_file_replacements[
                replace_replacements(tree_replacements, replacements, key)
            ] = file_replacements[key]
    else:
        replaced_tree = tree_lines
        replaced_file_replacements = file_replacements
    try:
        structure = TreeParser(replaced_tree)()
    except IndexError as idx:
        logger.error("Failed processing the tree. Tree looks like:")
        for line in replaced_tree:
            logger.error(line)
        logger.warning(
            "Please make sure your tree has no additional spaces, spacing is important. Full exception follows"
        )
        logger.warning(idx, exc_info=True)
        sys.exit(-1)
    logger.debug(replaced_file_replacements)
    structure_filler(structure, replaced_file_contents, replaced_file_replacements)
    return structure


def process_markdown(path: Path, replacements):
    """Convert markdown into a structure"""
    with open(path) as markdown_path:
        all_lines = markdown_path.read()
    return _process_markdown(all_lines, replacements)


def _process_markdown(all_lines, replacements):
    markdown, _ = MARKDOWN_PARSER.parse(all_lines)
    return parse_markdown_to_structure(markdown, replacements)


def materialise_structure(
    structure: Folder,
    path: Path,
    dry_run=True,
    replacements=None,
    ignore_existing_folders=False,
):
    """Materialise or dry run the folder structure"""
    logger.debug("Replacements: %s", replacements)
    location = path / Path(structure.as_posix())
    logger.info("🗃  %s", location)
    if not dry_run:
        try:
            location.mkdir()
        except FileExistsError:
            if ignore_existing_folders:
                logger.warning(
                    "Folder %s already exists and will not be recreated", location
                )
            else:
                logger.error(
                    "The destination folder %s already exists. Either delete it, use a new path or pass the --ignore-existing-folders flag",
                    location,
                )
                sys.exit(-1)

    for item in structure.iterdir():
        if item.is_dir():
            materialise_structure(item, path, dry_run, replacements, ignore_existing_folders)
        else:
            location = path / Path(item.as_posix())

            logger.info(
                "📁 %s (%s characters)", location, len("".join(item.contents)),
            )
            if replacements is not None and item.replacements is not None:
                new_contents = []
                for line in item.contents:
                    new_contents += [
                        replace_replacements(item.replacements, replacements, line)
                    ]
                item.set_contents("\n".join(new_contents))
                logger.info("Replacements applied to %s", item.name)
            if not dry_run:
                if location.exists():
                    logger.error(
                        "File %s already exists at that location. Delete it first",
                        location,
                    )
                    sys.exit(-1)
                with location.open("w") as destination:
                    for line in item.contents:
                        destination.write(line)
                        destination.write("\n")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['motllo']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.1,<8.0.0', 'colorlog>=4.1.0,<5.0.0', 'pyyaml>=5.3,<6.0']

entry_points = \
{'console_scripts': ['motllo = motllo.main:cli']}

setup_kwargs = {
    'name': 'motllo',
    'version': '0.2.2',
    'description': 'Project patterns like I like them',
    'long_description': '# Motllo\n\n[![PyPI version](https://badge.fury.io/py/motllo.svg)](https://badge.fury.io/py/motllo)\n\nRepository templates just like I like them\n\n- [Quick visual example](#quick-visual-example)\n- [Motivation](#motivation)\n- [Description](#description)\n- [Example](#example)\n- [Usage](#usage)\n- [Is it safe to use?](#is-it-safe-to-use)\n- [Next steps](#next-steps)\n\n## Quick visual example\n\n<img src="motllo.gif"/>\n\n## Motivation\n\nIt can be due to a bit of OCD and a bit of Knuth. \n\n- Most "project template" systems (that let you create a repository with\n  everything set up to use a library) need a specific repository for it\n  ([cookiecutter](https://cookiecutter.readthedocs.io/en/1.7.2/),\n  [giter8](http://www.foundweekends.org/giter8/), [GitHub\'s own template\n  repositories](https://help.github.com/en/github/creating-cloning-and-archiving-repositories/creating-a-template-repository)).\n  This feels weird to me, like needing a house to build a new house.\n- I prefer more compact stuff I can read. Even better if I can document\n\nSo, here\'s `motllo`\n\n```\n[1284; del ll. mŏdŭlus \'mesura; mòdul\', dimin. de modus \'mesura\']\n \t\nm 1 1 TECNOL METAL·L ART Peça amb una cavitat en la qual hom introdueix \nuna substància en forma de pols, de pasta o líquida per tal que, en \npassar a l\'estat sòlid, agafi la forma de la cavitat. \n```\n\n## Description\n\nMotllo can convert a folder structure/repository (it more or less handles\n`.gitignore` rules) into a Markdown document with:\n\n- A "Tree structure" section\n- Optionally, replacements to apply to the structure via the command line tool\n- A tree representation of the folder structure you want (similar to the one\n  generated by the `tree` UNIX utility, but in Python, thanks to a\n  [Stackoverflow answer](https://stackoverflow.com/a/59109706) by [Aaron\n  Hall](https://twitter.com/aaronchall)) inside of a code block\n- Optionally, section links\n- Sections, titled with the filepath described in the tree\n- Optionally, replacement blocks for this filepath\n- Code blocks (with possibly text describing what they are there for). All code\n  blocks within a section are merged into that file\n  \nAnd, probably more importantly, given a document as above, it can convert it into a\nfolder/repository structure, with variable replacement (in filenames as well as\nin code).\n\n## Example\n\nYou can see an example of a basic Python CLI tool (like this one) as a template,\nwith replacements and comments [here](examples/python_cli.md).\n\nYou can convert that Markdown file into a folder structure with\n\n```\nmotllo build PATH_TO/python_cli.md -o SOME_COOL_PATH -r "project_name:awesome-new-project --commit"\n```\n\nWithout `--commit` you will only get an informational log about what operations will be performed\n\nI plan on adding more examples at some point. New examples welcome!\n\n## Usage\n\nThe most easy uses are\n\n```\nmotllo markdown . -o some_markdown.md\n```\n\nto generate the Markdown template from a repository, which will respect nested\nand global `.gitignore ` and then\n\n```\nmotllo build markdown_template.md -o /wherever/ -r "project:cool-new-project" -r "version:0.0.1"\n```\n\nHere are the commands as shown in the CLI\n\n```\nUsage: motllo [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  --debug  Set log level to debug\n  --help   Show this message and exit.\n\nCommands:\n  build     Build a file/folder structure based on a Markdown document at...\n  markdown  Generate a Markdown template from a folder or repository at PATH\n  tree      Generate only the visual folder tree (like the UNIX tree...\n```\n---\n```\nUsage: motllo build [OPTIONS] PATH\n\n  Build a file/folder structure based on a Markdown document at PATH\n\nOptions:\n  -o, --output TEXT          Destination path to create everything  [required]\n  -r, --replace TEXT         Multiple replacement rules separated by colons,\n                             like -r "$PROJ:world_domination",\n                             -r"$TOOLS:python"\n\n  --ignore-existing-folders  Ignore if the destination folder already exists\n  --dry-run / --commit       Dry run by default so you can see what it does\n  --help                     Show this message and exit.\n```\n---\n```\nUsage: motllo markdown [OPTIONS] PATH\n\n  Generate a Markdown template from a folder or repository at PATH. Will\n  ignore hidden files, you can use --force-include to add them\n\nOptions:\n  -x, --max-length INTEGER      Maximum amount of lines to write in the\n                                markdown, for each file. Use -1 for `all of\n                                them`. Defaults to 15\n\n  -o, --output TEXT             Destination markdown file  [required]\n  --force-include TEXT          Glob patterns to forcefully include, comma\n                                separated between quotes like\n                                "*.py,*.c,*.scala"\n\n  --ignore TEXT                 Glob patterns to ignore, comma separated\n                                between quotes like "*.py,*.c,*.scala"\n\n  --gitignore / --no-gitignore  Use local and global gitignores, yes by\n                                default\n\n  --help                        Show this message and exit.\n```\n---\n```\nUsage: motllo tree [OPTIONS] PATH\n\n  Generate only the visual folder tree (like the UNIX tree command)\n\nOptions:\n  --force-include TEXT          Glob patterns to forcefully include, comma\n                                separated between quotes like\n                                "*.py,*.c,*.scala"\n\n  --ignore TEXT                 Glob patterns to ignore, comma separated\n                                between quotes like "*.py,*.c,*.scala"\n\n  --gitignore / --no-gitignore  Use local and global gitignores, yes by\n                                default\n\n  --help                        Show this message and exit.\n```\n\n## Is it safe to use?\n\nWell, in dry run mode it will tell you what it will write, so pretty much yes.\nDon\'t blame me if it fails. Of course, I\'d be happy to see it fail in the wild\nand improve it.\n\n## Next steps\n\nI want to refactor big chunks of the internals (I don\'t like parts of the API),\nand add some more testing, particularly around the replacements. This will be\ntricky, since generating random replacements is hard.\n\nI will probably spin out the markdown parser (which was stolen from the one I\nwrote for [bear-note-graph](https://github.com/rberenguel/bear-note-graph)) into\nits own library\n',
    'author': 'Ruben Berenguel',
    'author_email': 'ruben+poetry@mostlymaths.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/rberenguel/motllo',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
